@@ -31,16 +31,26 @@ try:
 except ImportError:
 	HAVE_SOCKS = False
 
+
+cchardet = False
+
+try:
+	import cchardet
+except ImportError:    # pragma: no cover
+	pass
+
+
 from . import HeaderParseMonkeyPatch
 
-from . import ChromiumMixin
 from . import Handlers
 from . import iri2uri
-from . import Constants
+from . import UA_Constants
+from . import Domain_Constants
 from . import Exceptions
 from . import utility
+from . import CloudscraperMixin
+from . import ChromiumMixin
 
-from .SeleniumModules import SeleniumPhantomJSMixin
 from .SeleniumModules import SeleniumChromiumMixin
 
 #pylint: disable-msg=E1101, C0325, R0201, W0702, W0703
@@ -56,10 +66,9 @@ GLOBAL_COOKIE_FILE = None
 # Arrrgh.
 
 class WebGetRobust(
-		ChromiumMixin.WebGetCrMixin,
-		SeleniumPhantomJSMixin.WebGetSeleniumPjsMixin,
-		SeleniumChromiumMixin.WebGetSeleniumChromiumMixin,
-
+			ChromiumMixin.WebGetCrMixin,
+			SeleniumChromiumMixin.WebGetSeleniumChromiumMixin,
+			CloudscraperMixin.WebGetCloudscraperMixin,
 		):
 
 	COOKIEFILE = 'cookies.lwp'				# the path and filename to save your cookies in
@@ -67,7 +76,8 @@ class WebGetRobust(
 	cookielib = None
 	opener = None
 
-	errorOutCount = 2
+	timeout = 30
+	errorOutCount = 1
 	# retryDelay = 0.1
 	retryDelay = 0.01
 
@@ -111,7 +121,7 @@ class WebGetRobust(
 		else:
 			# Due to general internet people douchebaggyness, I've basically said to hell with it and decided to spoof a whole assortment of browsers
 			# It should keep people from blocking this scraper *too* easily
-			self.browserHeaders = Constants.getUserAgent()
+			self.browserHeaders = UA_Constants.getUserAgent()
 
 		self.data = urllib.parse.urlencode(self.browserHeaders)
 
@@ -127,6 +137,56 @@ class WebGetRobust(
 		self.alt_cookiejar = alt_cookiejar
 		self.__loadCookies()
 
+
+
+	def _unwaf_func(self, funcname, requestedUrl, *args, **kwargs):
+		'''
+		Call func `funcname` on the WebRequest class, handling a WAF exception
+		automatically if it occurs
+		'''
+		target_func = getattr(self, funcname)
+
+		try:
+			return target_func(requestedUrl, *args, **kwargs)
+
+		except Exceptions.CloudFlareWrapper:
+			if self.rules['auto_waf']:
+				self.log.warning("Cloudflare failure! Doing automatic step-through.")
+				if not self.stepThroughCloudFlareWaf(requestedUrl):
+					raise Exceptions.FetchFailureError("Could not step through cloudflare!", requestedUrl)
+				# Cloudflare cookie set, retrieve again
+				return target_func(requestedUrl, *args, **kwargs)
+
+			else:
+				self.log.info("Cloudflare without step-through setting!")
+				raise
+
+		except Exceptions.SucuriWrapper:
+			# print("Sucuri!")
+			if self.rules['auto_waf']:
+				self.log.warning("Sucuri failure! Doing automatic step-through.")
+				if not self.stepThroughSucuriWaf(requestedUrl):
+					raise Exceptions.FetchFailureError("Could not step through Sucuri WAF bullshit!", requestedUrl)
+				return target_func(requestedUrl, *args, **kwargs)
+			else:
+				self.log.info("Sucuri without step-through setting!")
+				raise
+
+
+
+
+	def getpage(self, requestedUrl, *args, **kwargs):
+		'''
+		Get page at `requestedUrl`, while automatically handling a WAF,
+		if one is encountered
+
+		'''
+
+		return self._unwaf_func("_getpage", requestedUrl, *args, **kwargs)
+
+
+
+
 	def chunkReport(self, bytesSoFar, totalSize):
 		if totalSize:
 			percent = float(bytesSoFar) / totalSize
@@ -195,7 +255,7 @@ class WebGetRobust(
 					self.log.error("Retrying!")
 
 					# Scramble our current UA
-					self.browserHeaders = Constants.getUserAgent()
+					self.browserHeaders = UA_Constants.getUserAgent()
 					if self.alt_cookiejar:
 						self.cj.init_agent(new_headers=self.browserHeaders)
 
@@ -208,6 +268,35 @@ class WebGetRobust(
 					# 	tmp_err_fp.write(page)
 					raise
 
+
+
+	def getSoupNoRedirects(self, *args, **kwargs):
+		if 'returnMultiple' in kwargs:
+			raise Exceptions.ArgumentError("getSoup cannot be called with 'returnMultiple'")
+
+		if 'soup' in kwargs and kwargs['soup']:
+			raise Exceptions.ArgumentError("getSoup contradicts the 'soup' directive!")
+
+		kwargs['returnMultiple'] = True
+
+		tgt_url = kwargs.get('requestedUrl', None)
+		if not tgt_url:
+			tgt_url = args[0]
+
+
+		page, handle = self.getpage(*args, **kwargs)
+
+		redirurl = handle.geturl()
+		if redirurl != tgt_url:
+			self.log.error("Requested %s, redirected to %s. Raising error", tgt_url, redirurl)
+
+			raise Exceptions.RedirectedError("Requested %s, redirected to %s" % (
+				tgt_url, redirurl))
+
+		soup = as_soup(page)
+		return soup
+
+
 	def getFileAndName(self, *args, **kwargs):
 		'''
 		Give a requested page (note: the arguments for this call are forwarded to getpage()),
@@ -265,7 +354,132 @@ class WebGetRobust(
 
 		return pgctnt, hName, mime
 
-	def __getpage(self, requestedUrl, **kwargs):
+
+
+	def getItem(self, itemUrl):
+		content, handle = self.getpage(itemUrl, returnMultiple=True)
+
+		if not content or not handle:
+			raise urllib.error.URLError("Failed to retreive file from page '%s'!" % itemUrl)
+
+		handle_info = handle.info()
+
+		if handle_info['Content-Disposition'] and 'filename=' in handle_info['Content-Disposition'].lower():
+			fileN = handle_info['Content-Disposition'].split("=", 1)[-1]
+		else:
+			fileN = urllib.parse.unquote(urllib.parse.urlparse(handle.geturl())[2].split("/")[-1])
+			fileN = bs4.UnicodeDammit(fileN).unicode_markup
+		mType = handle_info['Content-Type']
+
+		# If there is an encoding in the content-type (or any other info), strip it out.
+		# We don't care about the encoding, since WebFunctions will already have handled that,
+		# and returned a decoded unicode object.
+		if mType and ";" in mType:
+			mType = mType.split(";")[0].strip()
+
+		# *sigh*. So minus.com is fucking up their http headers, and apparently urlencoding the
+		# mime type, because apparently they're shit at things.
+		# Anyways, fix that.
+		if mType and '%2F' in  mType:
+			mType = mType.replace('%2F', '/')
+
+		self.log.info("Retreived file of type '%s', name of '%s' with a size of %0.3f K", mType, fileN, len(content)/1000.0)
+		return content, fileN, mType
+
+	def getHead(self, url, addlHeaders=None):
+		self.log.warning("TODO: Fixme this neds to be migrated to use the normal fetch interface, so it is WAF-aware.")
+		for x in range(9999):
+			try:
+				self.log.info("Doing HTTP HEAD request for '%s'", url)
+				pgreq = self.__buildRequest(url, None, addlHeaders, None, req_class=Handlers.HeadRequest)
+				pghandle = self.opener.open(pgreq, timeout=self.timeout)
+				returl = pghandle.geturl()
+				if returl != url:
+					self.log.info("HEAD request returned a different URL '%s'", returl)
+
+				return returl
+			except socket.timeout as e:
+				self.log.info("Timeout, retrying....")
+				if x >= 3:
+					self.log.error("Failure fetching: %s", url)
+					raise Exceptions.FetchFailureError("Timout when fetching content", url)
+			except urllib.error.URLError as e:
+				# Continue even in the face of cloudflare crapping it's pants
+				if e.code == 500 and e.geturl():
+					return e.geturl()
+				self.log.info("URLError, retrying....")
+				if x >= 3:
+					self.log.error("Failure fetching: %s", url)
+					raise Exceptions.FetchFailureError("URLError when fetching content", e.geturl(), err_code=e.code)
+
+	######################################################################################################################################################
+	######################################################################################################################################################
+
+
+	def __check_suc_cookie(self, components):
+		'''
+		This is only called if we're on a known sucuri-"protected" site.
+		As such, if we do *not* have a sucuri cloudproxy cookie, we can assume we need to
+		do the normal WAF step-through.
+		'''
+		netloc = components.netloc.lower()
+
+		for cookie in self.cj:
+			if cookie.domain_specified and (cookie.domain.lower().endswith(netloc)
+				or (cookie.domain.lower().endswith("127.0.0.1") and (
+				components.path == "/sucuri_shit_3" or components.path == "/sucuri_shit_2" ))):   # Allow testing
+				if "sucuri_cloudproxy_uuid_" in cookie.name:
+					return
+		self.log.info("Missing cloudproxy cookie for known sucuri wrapped site. Doing a pre-emptive chromium fetch.")
+		raise Exceptions.SucuriWrapper("WAF Shit", str(components))
+
+	def __check_cf_cookie(self, components):
+		netloc = components.netloc.lower()
+
+		# TODO: Implement me?
+
+		# for cookie in self.cj:
+		# 	if cookie.domain_specified and (cookie.domain.lower().endswith(netloc)
+		# 		or (cookie.domain.lower().endswith("127.0.0.1") and components.path == "/sucuri_shit_2")):   # Allow testing
+
+		# 		if "sucuri_cloudproxy_uuid_" in cookie.name:
+		# 			return
+		# 		print("Target cookie!")
+		# 		print("K -> V: %s -> %s" % (cookie.name, cookie.value))
+
+		# 	print(cookie)
+		# 	print(type(cookie))
+		# 	print(cookie.domain)
+		# raise RuntimeError
+		pass
+
+	def __pre_check(self, requestedUrl):
+		'''
+		Allow the pre-emptive fetching of sites with a full browser if they're known
+		to be dick hosters.
+		'''
+		components = urllib.parse.urlsplit(requestedUrl)
+
+		netloc_l = components.netloc.lower()
+		if netloc_l in Domain_Constants.SUCURI_GARBAGE_SITE_NETLOCS:
+			self.__check_suc_cookie(components)
+		elif netloc_l in Domain_Constants.CF_GARBAGE_SITE_NETLOCS:
+			self.__check_cf_cookie(components)
+
+		# Testing
+		elif components.path == '/sucuri_shit_2':
+			self.__check_suc_cookie(components)
+		elif components.path == '/sucuri_shit_3':
+			self.__check_suc_cookie(components)
+		elif components.path == '/cloudflare_under_attack_shit_2':
+			self.__check_cf_cookie(components)
+		elif components.path == '/cloudflare_under_attack_shit_3':
+			self.__check_cf_cookie(components)
+
+
+	def _getpage(self, requestedUrl, **kwargs):
+		self.__pre_check(requestedUrl)
+
 		self.log.info("Fetching content at URL: %s", requestedUrl)
 
 		# strip trailing and leading spaces.
@@ -323,7 +537,7 @@ class WebGetRobust(
 			#print "execution", retryCount
 			try:
 				# print("Getpage!", requestedUrl, kwargs)
-				pghandle = self.opener.open(pgreq, timeout=30)					# Get Webpage
+				pghandle = self.opener.open(pgreq, timeout=self.timeout)					# Get Webpage
 				# print("Gotpage")
 
 			except Exceptions.GarbageSiteWrapper as err:
@@ -356,15 +570,16 @@ class WebGetRobust(
 					break
 
 				time.sleep(self.retryDelay)
-				if err.code == 503:
-					if err_content and b'This process is automatic. Your browser will redirect to your requested content shortly.' in err_content:
-						raise Exceptions.CloudFlareWrapper("WAF Shit", requestedUrl)
+
+				if (err.code == 403 or err.code == 429 or err.code == 503) and err_content:
+
+					self._check_waf(err_content, requestedUrl)
 
 				# So I've been seeing this causing CF to bounce too.
 				# As such, poke through those via chromium too.
 				if err.code == 502:
-					if err_content and b'is currently offline. However, because the site uses Cloudflare\'s Always Online' in err_content:
-						raise Exceptions.CloudFlareWrapper("WAF Shit", requestedUrl)
+					if err_content:
+						self._check_waf(err_content, requestedUrl)
 
 			except UnicodeEncodeError:
 				self.log.critical("Unrecoverable Unicode issue retrieving page - %s", requestedUrl)
@@ -428,92 +643,10 @@ class WebGetRobust(
 		else:
 			return pgctnt
 
-	def getpage(self, requestedUrl, *args, **kwargs):
-		try:
-			return self.__getpage(requestedUrl, *args, **kwargs)
-
-		except Exceptions.CloudFlareWrapper:
-			if self.rules['auto_waf']:
-				self.log.warning("Cloudflare failure! Doing automatic step-through.")
-				if not self.stepThroughCloudFlareWaf(requestedUrl):
-					raise Exceptions.FetchFailureError("Could not step through cloudflare!", requestedUrl)
-				# Cloudflare cookie set, retrieve again
-				return self.__getpage(requestedUrl, *args, **kwargs)
-
-			else:
-				self.log.info("Cloudflare without step-through setting!")
-				raise
-
-		except Exceptions.SucuriWrapper:
-			# print("Sucuri!")
-			if self.rules['auto_waf']:
-				self.log.warning("Sucuri failure! Doing automatic step-through.")
-				if not self.stepThroughSucuriWaf(requestedUrl):
-					raise Exceptions.FetchFailureError("Could not step through Sucuri WAF bullshit!", requestedUrl)
-				return self.__getpage(requestedUrl, *args, **kwargs)
-			else:
-				self.log.info("Sucuri without step-through setting!")
-				raise
-
-	def getItem(self, itemUrl):
-		content, handle = self.getpage(itemUrl, returnMultiple=True)
-
-		if not content or not handle:
-			raise urllib.error.URLError("Failed to retreive file from page '%s'!" % itemUrl)
-
-		handle_info = handle.info()
-
-		if handle_info['Content-Disposition'] and 'filename=' in handle_info['Content-Disposition'].lower():
-			fileN = handle_info['Content-Disposition'].split("=", 1)[-1]
-		else:
-			fileN = urllib.parse.unquote(urllib.parse.urlparse(handle.geturl())[2].split("/")[-1])
-			fileN = bs4.UnicodeDammit(fileN).unicode_markup
-		mType = handle_info['Content-Type']
-
-		# If there is an encoding in the content-type (or any other info), strip it out.
-		# We don't care about the encoding, since WebFunctions will already have handled that,
-		# and returned a decoded unicode object.
-		if mType and ";" in mType:
-			mType = mType.split(";")[0].strip()
-
-		# *sigh*. So minus.com is fucking up their http headers, and apparently urlencoding the
-		# mime type, because apparently they're shit at things.
-		# Anyways, fix that.
-		if mType and '%2F' in  mType:
-			mType = mType.replace('%2F', '/')
-
-		self.log.info("Retreived file of type '%s', name of '%s' with a size of %0.3f K", mType, fileN, len(content)/1000.0)
-		return content, fileN, mType
-
-	def getHead(self, url, addlHeaders=None):
-		for x in range(9999):
-			try:
-				self.log.info("Doing HTTP HEAD request for '%s'", url)
-				pgreq = self.__buildRequest(url, None, addlHeaders, None, req_class=Handlers.HeadRequest)
-				pghandle = self.opener.open(pgreq, timeout=30)
-				returl = pghandle.geturl()
-				if returl != url:
-					self.log.info("HEAD request returned a different URL '%s'", returl)
-
-				return returl
-			except socket.timeout as e:
-				self.log.info("Timeout, retrying....")
-				if x >= 3:
-					self.log.error("Failure fetching: %s", url)
-					raise Exceptions.FetchFailureError("Timout when fetching content", url)
-			except urllib.error.URLError as e:
-				# Continue even in the face of cloudflare crapping it's pants
-				if e.code == 500 and e.geturl():
-					return e.geturl()
-				self.log.info("URLError, retrying....")
-				if x >= 3:
-					self.log.error("Failure fetching: %s", url)
-					raise Exceptions.FetchFailureError("URLError when fetching content", e.geturl(), err_code=e.code)
-
 	######################################################################################################################################################
 	######################################################################################################################################################
 
-	def __decodeHtml(self, pageContent, cType):
+	def __decode_text_content(self, pageContent, cType):
 
 		# this *should* probably be done using a parser.
 		# However, it seems to be grossly overkill to shove the whole page (which can be quite large) through a parser just to pull out a tag that
@@ -522,7 +655,7 @@ class WebGetRobust(
 
 		# Regex is of bytes type, since we can't convert a string to unicode until we know the encoding the
 		# bytes string is using, and we need the regex to get that encoding
-		coding = re.search(rb"charset=[\'\"]?([a-zA-Z0-9\-]*)[\'\"]?", pageContent, flags=re.IGNORECASE)
+		coding = re.search(b"charset=[\'\"]?([a-zA-Z0-9\-]*)[\'\"]?", pageContent, flags=re.IGNORECASE)
 
 		cType = b""
 		charset = None
@@ -540,9 +673,23 @@ class WebGetRobust(
 				dummy_docType, charset = cType.split(b";")
 				charset = charset.split(b"=")[-1]
 
+		if cchardet:
+			inferred = cchardet.detect(pageContent)
+			if inferred and inferred['confidence'] is None:
+				# If we couldn't infer a charset, just short circuit and return the content.
+				# It's probably binary.
+				return pageContent
+
+			elif inferred and inferred['confidence'] is not None and inferred['confidence'] > 0.8:
+				charset = inferred['encoding']
+				self.log.info("Cchardet inferred encoding: %s", charset)
+
+		else:
+			self.log.warning("Missing cchardet!")
+
 		if not charset:
 			self.log.warning("Could not find encoding information on page - Using default charset. Shit may break!")
-			charset = "iso-8859-1"
+			charset = "utf-8"
 
 		try:
 			pageContent = str(pageContent, charset)
@@ -591,6 +738,9 @@ class WebGetRobust(
 			raise
 
 	def __decompressContent(self, coding, pgctnt):
+		"""
+		This is really obnoxious
+		"""
 		#preLen = len(pgctnt)
 		if coding == 'deflate':
 			compType = "deflate"
@@ -635,7 +785,7 @@ class WebGetRobust(
 			pgctnt = f.read()
 
 		elif coding == "sdch":
-			raise Exceptions.ContentTypeError("Wait, someone other then google actually supports SDCH compression?", pgreq)
+			raise Exceptions.ContentTypeError("Wait, someone other then google actually supports SDCH compression (%s)?" % pgreq)
 
 		else:
 			compType = "none"
@@ -669,20 +819,14 @@ class WebGetRobust(
 				# Use content-aware mechanisms for determing the content encoding.
 
 
-				if "text/html" in cType or \
-					'text/javascript' in cType or    \
-					'text/css' in cType or    \
-					'application/xml' in cType or    \
-					'application/atom+xml' in cType:				# If this is a html/text page, we want to decode it using the local encoding
-
-					pgctnt = self.__decodeHtml(pgctnt, cType)
-
-				elif "text/plain" in cType or "text/xml" in cType:
-					pgctnt = bs4.UnicodeDammit(pgctnt).unicode_markup
-
-				# Assume JSON is utf-8. Probably a bad idea?
-				elif "application/json" in cType:
-					pgctnt = pgctnt.decode('utf-8')
+				if "text/html" in cType or             \
+					'text/javascript' in cType or      \
+					'text/css' in cType or             \
+					'application/json' in cType or     \
+					'application/xml' in cType or      \
+					'application/atom+xml' in cType or \
+					cType.startswith("text/"):				# If this is a html/text page, we want to decode it using the local encoding
+					pgctnt = self.__decode_text_content(pgctnt, cType)
 
 				elif "text" in cType:
 					self.log.critical("Unknown content type!")
@@ -723,8 +867,8 @@ class WebGetRobust(
 			else:
 				self.log.info("Compression type = %s. Content Size compressed = %0.3fK. Decompressed = %0.3fK. File type: %s.", compType, preDecompSize, decompSize, cType)
 
-			if b"sucuri_cloudproxy_js=" in pgctnt:
-				raise Exceptions.SucuriWrapper("WAF Shit", pgreq)
+			self._check_waf(pgctnt, pgreq.get_full_url())
+
 			pgctnt = self.__decodeTextContent(pgctnt, cType)
 
 			return pgctnt
@@ -756,6 +900,19 @@ class WebGetRobust(
 		# postData expects a dict
 		# addlHeaders also expects a dict
 
+	def _check_waf(self, pageContent, pageUrl):
+		assert isinstance(pageContent, bytes), "Item pageContent must be of type bytes, received %s" % (type(pageContent), )
+		assert isinstance(pageUrl, str), "Item pageUrl must be of type str, received %s" % (type(pageUrl), )
+
+		if b"sucuri_cloudproxy_js=" in pageContent:
+			raise Exceptions.SucuriWrapper("WAF Shit", pageUrl)
+
+		if b'This process is automatic. Your browser will redirect to your requested content shortly.' in pageContent:
+			raise Exceptions.CloudFlareWrapper("WAF Shit", pageUrl)
+
+		if b'is currently offline. However, because the site uses Cloudflare\'s Always Online' in pageContent:
+			raise Exceptions.CloudFlareWrapper("WAF Shit", pageUrl)
+
 	######################################################################################################################################################
 	######################################################################################################################################################
 
@@ -946,7 +1103,9 @@ class WebGetRobust(
 
 
 	def stepThroughCloudFlareWaf(self, url):
-		return self.stepThroughJsWaf(url, titleNotContains='Just a moment...')
+		# return self.stepThroughJsWaf(url, titleNotContains='Just a moment...')
+		return self.handle_cloudflare_cloudscraper(url)
+
 	def stepThroughSucuriWaf(self, url):
 		return self.stepThroughJsWaf(url, titleNotContains="You are being redirected...")
 
@@ -958,3 +1117,14 @@ class WebGetRobust(
 	# Compat for old code.
 	def stepThroughCloudFlare(self, *args, **kwargs):
 		return self.stepThroughJsWaf(*args, **kwargs)
+
+
+class PlainWafWebGetRobust(WebGetRobust):
+	def stepThroughCloudFlareWaf(self, url):
+		return self.stepThroughJsWaf(url, titleNotContains='Just a moment...')
+
+
+if __name__ == '__main__':
+	cs = WebGetRobust()
+	cs.handle_cloudflare_cloudscraper("https://www.cloudflare.com/")
+
